
/* 
 * dnstest.c -- test DNS connectivity past firewalls
 * Copyright (c) Jarkko Turkulainen 2004. All rights reserved.
 *
 * Platforms: Windows 2K/XP
 *
 * Compile: cl dnstest.c
 *
 * Parts of this code are based on "DNShell" (http://www.klake.org/~jt/dnshell/)
 *
 */


#include <stdio.h>
#include <winsock2.h>
#include <windows.h>

#pragma comment(lib, "ws2_32.lib")
#pragma comment(lib, "advapi32.lib")


/* DNS query template */

unsigned char dns_template[] = 
"\xd0\x0d"				/* Transaction ID (d00d) */
"\x01\x00"				/* Flags (recursive query) */
"\x00\x01\x00\x00\x00\x00\x00\x00"	/* 1 question section */

/* dns-encoded www.microsoft.com */
"\x03"					/* 3 bytes follow (www) */
"www"
"\x09"					/* 9 bytes follow (microsoft) */
"microsoft"
"\x03"					/* 3 bytes follow (com) */
"com"
"\x00"					/* name end */

"\x00\x01\x00\x01";			/* Type: A, Class: Inet */


char msg_caption[] = "DNStest v1.0 - (c) Jarkko Turkulainen 2004.";

char msg_success[] =
	"\nYour computer has just made a successful recursive DNS query\n"
	"for www.microsoft.com using svchost.exe as a host process.\n"
	"This means that it is possible to transfer information from your\n"
	"computer past personal and network firewalls.\n\n";

char msg_error_net[] =
	"\nThere seems to be some problem with your network setup or\n"
	"your personal firewall software has blocked this attempt.\n\n";

char msg_error_dns[] =
	"\nBecause you see this message, your personal firewall software\n"
	"hasn't been able to block the data transfer. However, there\n"
	"seems to be some problem with the DNS server. Try again later.\n\n";



/* Return values */
#define	DNS_RESPONSE_OK		0
#define	DNS_RESPONSE_FORMAT	1
#define	DNS_RESPONSE_SERV_FAIL	2
#define	DNS_RESPONSE_NAME_ERROR	3
#define	DNS_RESPONSE_NOT_IMPL	4
#define	DNS_RESPONSE_REFUSED	5
#define	DNS_RESPONSE_NETWORK	10
#define	DNS_RESPONSE_TIMEOUT	100


/* DNS defines */
#define DNS_MAX_PKTSIZE		512
#define DNS_PORT		53
#define DNS_TIMEOUT		2
#define DNS_POLL_TIMEOUT	5000
#define DNS_MIN_SIZE		30
#define DNS_HEADER_SIZE		12
#define DNS_TYPE_SIZE		2
#define DNS_CLASS_SIZE		2
#define DNS_LEN_SIZE		2
#define DNS_TTL_SIZE		4
#define DNS_TIMER_ID		0


/* Function prototypes */

typedef VOID (WINAPI *f_ExitProcess)(UINT);
typedef HINSTANCE (WINAPI *f_LoadLibrary)(LPCTSTR);
typedef UINT (WINAPI *f_WinExec)(LPCSTR, UINT);
typedef int (WINAPI *f_WSAStartup) (WORD, LPWSADATA);
typedef SOCKET (WINAPI *f_socket)(int, int, int);
typedef int (WINAPI *f_select)(int, fd_set *, fd_set *, fd_set *, const struct timeval *);
typedef int (WINAPI *f_sendto)(SOCKET, const char *, int, int, const struct sockaddr *, int);
typedef int (WINAPI *f_recvfrom)(SOCKET, char *, int, int, struct sockaddr *, int *);
typedef int (WINAPI *f_closesocket)(SOCKET);
typedef int (WINAPI *f_MessageBox)(HWND, LPCTSTR, LPCTSTR, UINT);


/* Injected code context */

typedef struct _INJECT_CTX {

	/* Module names */
	char			user32[32];
	char			ws2_32[32];
	/* Function VMAs */
	f_LoadLibrary		LoadLibrary;
	f_ExitProcess		ExitProcess;
	f_WinExec		WinExec;
	f_MessageBox		MessageBox;
	f_WSAStartup		WSAStartup;
	f_socket		socket;
	f_select		select;
	f_sendto		sendto;
	f_recvfrom		recvfrom;
	f_closesocket		closesocket;

	/* Networking structures */
	struct hostent 		*hp;
	struct sockaddr_in 	adr;

	/* DNS request */
	unsigned char		request[DNS_MAX_PKTSIZE];
	int			request_len;

	/* MessageBox data */
	char			caption[64];
	char			success[256];
	char			error_dns[256];
	char			error_net[256];

} INJECT_CTX;


void inject_start() {

	INJECT_CTX 	*data;
	WSADATA 	wsa_data;
	struct timeval	timeout;
	fd_set 		rs;
	char		resp[DNS_MAX_PKTSIZE] = {0};
	int		sd, c, slen, index, compression;

	/* 
	 * We have the data context pointer in cpu register esi
	 * All we need to do is to point esi to stack variable "data"
 	 * which we declared as a pointer to struct _INJECT_CTX.
 	 *
	 */
	__asm  {
		mov data, esi
	}

	/* Next, load all needed libraries */
	data->LoadLibrary(data->user32);
	data->LoadLibrary(data->ws2_32);

	/* Winsock startup routines */
	data->WSAStartup(MAKEWORD(2,0), &wsa_data);

	/* Timeout 10 secs */
	timeout.tv_sec  = 10;
	timeout.tv_usec = 0;

	/* Request socket */
	sd = data->socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

	/* Send request */
	slen = sizeof(data->adr);
	c = data->sendto(sd, data->request, data->request_len, 0, 
		(struct sockaddr *)&data->adr, slen);
	if (c == -1) {
		data->closesocket(sd);
		data->MessageBox(NULL, data->error_net, data->caption, MB_OK);
		data->ExitProcess(1);
        }

	/* Watch for socket read */
	FD_ZERO(&rs);
	FD_SET(sd, &rs);
	c = data->select((sd + 1), &rs, NULL, NULL, &timeout);

	/* Nothing interesting */
	if (c == 0) {
		data->closesocket(sd);
		data->MessageBox(NULL, data->error_net, data->caption, MB_OK);
		data->ExitProcess(1);
        }

	/* Get response */
	c = data->recvfrom(sd, resp, DNS_MAX_PKTSIZE, 0,
		(struct sockaddr *)&data->adr, &slen);
 	if (c == -1) {
		data->closesocket(sd);
		data->MessageBox(NULL, data->error_net, data->caption, MB_OK);
		data->ExitProcess(1);
        }

	/* Check response type */
	if ((resp[3]&0x0f) == DNS_RESPONSE_OK)
		data->MessageBox(NULL, data->success, data->caption, MB_OK);
	else
		data->MessageBox(NULL, data->error_dns, data->caption, MB_OK);

	data->closesocket(sd);
	data->ExitProcess(0);

}

void __declspec(naked) inject_end() {

	__asm ret
}


/* XXXXXXXXXXXXXXx horrible hack... need some better */
char *find_dns_server() {
	HKEY 			Result, Interface;
	LONG 			Return;
	DWORD 			Regtype, Size, Index;
	char			Data[1024], *ptr;
	PFILETIME		PFile;
	int			dlen;

	PFile = 0;
	Size = sizeof(Data);
	memset(Data, 0, Size);
	
	/* TCPIP Interface key */
	Return = RegOpenKeyEx(HKEY_LOCAL_MACHINE, 
		TEXT("SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Interfaces"), 
		0, KEY_READ, &Result);

	/* Success */
	if (Return == ERROR_SUCCESS) {
		Index = 0;

		/* Loop through all subkeys */
		while ((Return = RegEnumKeyEx(Result, Index, Data, &Size, 
				NULL, NULL, NULL, PFile)) != ERROR_NO_MORE_ITEMS) {
			RegOpenKeyEx(Result, Data, 0, KEY_READ, &Interface);

			/* Try "NameServer" value */
			Return = RegQueryValueEx(Interface, TEXT("NameServer"),
				NULL, &Regtype, Data, &Size);

			/* No success, try "DhcpNameServer" value instead */
			if ((Return != ERROR_SUCCESS) || (Size == 1)) {
				Size = 1024;
				memset(Data, 0, Size);
				Return = RegQueryValueEx(Interface, 
					TEXT("DhcpNameServer"),
					NULL, &Regtype, Data, &Size);
			}

			/* If success, try out the servers */
			if ((Return == ERROR_SUCCESS) && (Size > 1)) {
			
				/* Multiple servers are separated with comma (',') */
				dlen = strlen(Data);
				for (ptr = Data; *ptr != '\0'; ptr++) {
					if (*ptr == ',') *ptr = '\0';
				}

				/* Return first server */
				return(strdup(Data));
					
			}
			Size = 1024;
			memset(Data, 0, Size);

			Index++;
		}
	} 
	RegCloseKey(Result);

	return(NULL);

}


int WinMain (HINSTANCE hInst, HINSTANCE hPrev, LPSTR lpCmd, int nShow) {
	STARTUPINFO 		si = {0};
	PROCESS_INFORMATION 	pi = {0};
	INJECT_CTX		data = {0};
	CONTEXT 		ctx;
	LPVOID                  code_address, data_address;
	unsigned char 		*start, *stop;
	unsigned long		length;
	char			*dnsserver;
	WSADATA 		wsa_data;

	/* Module handles */
	LoadLibrary("user32");
	LoadLibrary("ws2_32");
	data.LoadLibrary = (f_LoadLibrary)GetProcAddress(GetModuleHandle("kernel32"), 
		"LoadLibraryA");
	data.ExitProcess = (f_ExitProcess)GetProcAddress(GetModuleHandle("kernel32"), 
		"ExitProcess");
	data.WinExec = (f_WinExec)GetProcAddress(GetModuleHandle("kernel32"), 
		"WinExec");
	data.WSAStartup = (f_WSAStartup)GetProcAddress(GetModuleHandle("ws2_32"), 
		"WSAStartup");
	data.socket = (f_socket)GetProcAddress(GetModuleHandle("ws2_32"), 
		"socket");
	data.closesocket = (f_closesocket)GetProcAddress(GetModuleHandle("ws2_32"), 
		"closesocket");
	data.select = (f_select)GetProcAddress(GetModuleHandle("ws2_32"), 
		"select");
	data.sendto = (f_sendto)GetProcAddress(GetModuleHandle("ws2_32"), 
		"sendto");
	data.recvfrom = (f_recvfrom)GetProcAddress(GetModuleHandle("ws2_32"), 
		"recvfrom");
	data.MessageBox = (f_MessageBox)GetProcAddress(GetModuleHandle("user32"), 
		"MessageBoxA");

	/* Initialize strings */
	lstrcpy(data.user32, "user32");
	lstrcpy(data.ws2_32, "ws2_32");
	lstrcpy(data.caption, msg_caption);
	lstrcpy(data.success, msg_success);
	lstrcpy(data.error_net, msg_error_net);
	lstrcpy(data.error_dns, msg_error_dns);
	memcpy(data.request, dns_template, sizeof(dns_template) - 1);
	data.request_len = sizeof(dns_template) - 1;
	
	/* Find out the DNS server from registry */
	dnsserver = (char *)find_dns_server();

	/* Winsock startup routines */
	WSAStartup(MAKEWORD(2,0), &wsa_data);

	/* Fill in the server socket structure */
	data.adr.sin_family=AF_INET;
	data.adr.sin_port=htons((short int)53);
	if((data.adr.sin_addr.s_addr = inet_addr(dnsserver)) == -1) {
		if ((data.hp = gethostbyname(dnsserver)) == NULL) {
			exit (1);
		}
		memcpy(&data.adr.sin_addr.s_addr, data.hp->h_addr, 4);
	}

	/* Create the host process */
	CreateProcess(NULL, "svchost.exe", NULL, NULL, 0,
		CREATE_SUSPENDED|CREATE_NO_WINDOW, NULL, NULL, &si, &pi);
	
	/* Get process context */
	ctx.ContextFlags = CONTEXT_FULL;
	GetThreadContext(pi.hThread, &ctx);
	
	/* Calculate injected code size */
	start = (unsigned char *)((unsigned char *)inject_start);
	stop  = (unsigned char *)((unsigned char *)inject_end);
	length = stop - start;

	/* Allocate space for injected code */
	code_address = VirtualAllocEx(pi.hProcess, NULL, length, 
		MEM_COMMIT, PAGE_EXECUTE_READWRITE);

	/* Inject code */
	WriteProcessMemory(pi.hProcess, code_address, start, length, NULL);

	/* Allocate space for code context */
	data_address = VirtualAllocEx(pi.hProcess, NULL, sizeof(data), 
		MEM_COMMIT, PAGE_EXECUTE_READWRITE);

	/* Write the context */
	WriteProcessMemory(pi.hProcess, data_address, (LPVOID)&data, 
		sizeof(data), NULL);

	/* Set process context */
	ctx.Esi = (unsigned long)data_address;
	ctx.Eip = (unsigned long)code_address;
	SetThreadContext(pi.hThread, &ctx);
	

	/* Start the resumed process */
	ResumeThread(pi.hThread);
	
	
	return (0);
}

